import signal
from ._version import version as __version__
import avtoolbox._av_base

__author__ = "Simulation Based Engineering Laboratory (negrut@.wisc.edu)"
"""Simulation Based Engineering Laboratory (negrut@wisc.edu)"""
__license__ = "BSD3"
"""BSD3"""

def _signal_handler(sig, frame):
    """Signal handler that will exit if ctrl+c is recorded in the terminal window.

    Allows easier exiting of a matplotlib plot

    Args:
        sig (int): Signal number
        frame (int): ?
    """

    import sys
    sys.exit(0)


# setup the signal listener to listen for the interrupt signal (ctrl+c)
signal.signal(signal.SIGINT, _signal_handler)

del signal




"""
Provides a utility class :class:`~MessageType` to aid in registering message types
"""

# General imports
from typing import NamedTuple

class MessageType(NamedTuple):
    """
    ROS bags typically include custom message which are difficult to read directly since the data itself doesn't include information about how it's structured. 
    When writing our own rosbags, we want to store this information to make it easier to parse later. To accomplish this, this class contains information necessary to
    register the message for writing, and then reading later.

    Attributes:
        file (str):     the path to the .msg file that defines the data types
        name (str):     the nameof the message type (i.e. std_msgs/msg/Header)
    """ 
    file: str
    name: str





"""
Parser for yaml files.

Yaml files are human readable configuration files: https://yaml.org/
"""

# Import some utilities
from avtoolbox.utils.logger import LOGGER
from avtoolbox.utils.files import file_exists, get_file_type

# External library imports
import yaml

class YAMLParser:
    def __init__(self, filename):
        # Do some checks first
        self._filename = filename
        if not file_exists(filename):
            self._data = {}
            return

        # Load in the file
        LOGGER.info(f"Reading {filename} as yaml...")
        with open(filename, "r") as f:
            self._data = yaml.safe_load(f)
        LOGGER.debug(f"Read {filename} as yaml.")

    def contains(self, *args) -> bool:
        """
        Checks whether the yaml file contains a certain nested attribute

        Ex:
            ```test.yml
            test:
                one: red
                two: blue
                three: green
            ```

            parser = YAMLParser('test.yml')
            parser.contains('test', 'one')          // true
            parser.contains('test', 'four')         // false
            paresr.contains('test', 'one', 'red')   // false; only will search keys
           
        Args:
            *args: A list of arguments to search in the file

        Returns:
            bool: Whether the nested attributes are contained in the file
        """
        LOGGER.debug(f"Checking if {self._filename} contains nested attributes: {args}...")

        # If no data is available, always return False:
        if self._data is None:
            return False

        _contains = True
        temp = self._data
        for arg in args:
            if arg not in temp:
                _contains = False
                LOGGER.info(f"{self._filename} does not contain nested attributes: {args}.")
                break
            temp = temp[arg]
        return _contains

    def get(self, *args, default=None, throw_error=True) -> 'Any':
        """
        Grabs the attribute at the nested location provided by args

        Ex:
            ```test.yml
            test:
                one: red
                two: blue
                three: green
            ```

            parser = YAMLParser('test.yml')
            paresr.get('test', 'one')               // red 
            paresr.get('test', 'green', 'test')     // test
            paresr.get('test', 'green')             // raises AttributeError
           
        Args:
            *args: A list of arguments to search in the file
            default (Any): The default value if the nested attribute isn't found
            throw_error (bool): Throw an error if default is None and the attribute isn't found. Defaults to True.

        Returns:
            Any: The value at the nested attributes

        Raises:
            KeyError: If the nested attributes don't actually point to a value (i.e. contains(args) == False)
        """
        LOGGER.debug(f"Getting nested attributes from {self._filename}: {args}...")

        temp = self._data
        if temp is not None:
            for arg in args:
                if arg not in temp:
                    LOGGER.info(f"{self._filename} does not contain nested attributes: {args}.")
                    if default is not None:
                        LOGGER.info(f"Using default: {default}.")
                    temp = default
                    break
                temp = temp[arg]

        if temp is None and throw_error:
            raise AttributeError(f"Default is not set and the nested attribute was not found: {args}.")

        return temp 



"""
Provides a custom logger object.

Should be used like the following:

.. highlight:: python
.. code-block:: python
    
    from avtoolbox.utils.logger import LOGGER

    LOGGER.fatal("Fatal")
    LOGGER.error("Error")
    LOGGER.warn("Warning")
    LOGGER.info("Information")
    LOGGER.debug("Debug")
"""

import logging
from colorlog import ColoredFormatter

# Create logger
LOGGER = logging.getLogger(__name__)

# Create a handler with the desired format
DEFAULT_LOGGING_FORMAT = '%(log_color)s%(levelname)-8s%(reset)s | %(log_color)s%(module)s.%(funcName)s :: %(message)s%(reset)s'
CONSOLE_HANDLER = logging.StreamHandler()
CONSOLE_HANDLER.setFormatter(ColoredFormatter(DEFAULT_LOGGING_FORMAT))

# Set default logging levels
DEFAULT_LOGGING_LEVEL = logging.WARNING
CONSOLE_HANDLER.setLevel(DEFAULT_LOGGING_LEVEL)
LOGGER.setLevel(DEFAULT_LOGGING_LEVEL)

# Add the handler to the logger
LOGGER.addHandler(CONSOLE_HANDLER)


def set_verbosity(verbosity: int):
    """
    Set the verbosity level for the logger.
    Default verbosity is WARNING. ``verbosity`` should be a value
    greater than 0 and less than 2 that represents how many levels below WARNING is desired.
    For instance, if ``verbosity`` is 1, the new level will be INFO because
    that is one below WARNING. If ``verbosity`` is 2, the new level is DEBUG.
    DEBUG and INFO are currently the only two implemented

    Args:
        verbosity (int): A value between 0 and 2 that represents the number of levels below WARNING that should be used when logging.
    """

    if verbosity < 0 or verbosity > 2:
        raise ValueError(
            f"Verbosity should be greater than 0 and less than 3. Got {verbosity}.")

    level = DEFAULT_LOGGING_LEVEL - verbosity * 10
    LOGGER.setLevel(level)
    CONSOLE_HANDLER.setLevel(level)
    LOGGER.log(
        level, f"Verbosity has been set to {logging.getLevelName(level)}")


def dumps_dict(dic: dict) -> str:
    """
    Dumps a dictionary in a pretty-ish format to the logger.

    Args:
       dic (dict): The dictionary to print 

    Returns:
        str: The pretty-ish string representation of the dict argument
    """
    import json
    return json.dumps(dic, indent=4)



"""
The entrypoint for the AV Toolkit CLI is `av`
The main parser will have a few commands, such as verbosity or a help menu.
For the most part, the entrypoint will be used to access subparsers,
such as `db` to interact with the AV database.
"""
# Command imports
import avtoolbox.db as db
import avtoolbox.dev as dev

# Utility imports
from avtoolbox.utils.logger import set_verbosity

# General imports
import argparse

def _init():
    """
    The root entrypoint for the AV CLI is `av`. This the first command you need to access the CLI. All subsequent subcommands succeed `av`.
    """
    # Main entrypoint and initialize the cmd method
    # set_defaults specifies a method that is called if that parser is used
    parser = argparse.ArgumentParser(description="AV Toolkit Command Line Interface")
    parser.add_argument('-v', '--verbose', dest='verbosity', action='count', help='Level of verbosity', default=0)
    parser.add_argument('--dry-run', action="store_true", help="Test run that will print out the commands it would run but not actually run them")
    parser.set_defaults(cmd=lambda x: x)

    # Initialize the subparsers
    subparsers = parser.add_subparsers()
    db._init(subparsers.add_parser("db", description="Interact with the niAV database"))  # noqa
    dev._init(subparsers.add_parser("dev", description="Work with the niAV development environment"))

    return parser

def _main():
    # Create the parser
    parser = _init()

    # Parse the arguments and update logging
    args = parser.parse_args()
    set_verbosity(args.verbosity)

    # Calls the cmd for the used subparser
    args.cmd(args)



"""
Provides helper methods for interacting with the filesystem.
"""

# Import some utils
from avtoolbox.utils.logger import LOGGER

# External library imports
from pathlib import Path
from typing import Union, Tuple
import shutil

def as_path(path: str) -> Path:
    """
    Simple helper method to get a path as a :class:`pathlib.Path` object

    Args:
        path: the path to convert

    Returns:
        Path: The :class:`pathlib.Path` object
    """
    return Path(path)

def copy_file(path: Union[Path, str], dest: Union[Path, str]):
    """
    Copy a file/directory (``path``) to some destination (``dest``).

    Args:
        path (Union[Path, str]): The file/directory source to copy.
        dest (Union[Path, str]): The file/directory destination to copy to.
    """
    shutil.copy(path, dest)

def file_exists(filename: str, throw_error: bool = False, can_be_directory: bool = False) -> bool:
    """
    Check if the passed filename is an actual file

    Args:
        filename (str): The filename to check
        throw_error (bool): If True, will throw an error if the file doesn't exist. Defaults to False.
        can_be_directory (bool): If True, will check if it is a directory, in addition to a file

    Returns:
        bool: True if the file exists, false otherwise

    Throws:
        FileNotFoundError: If filename is not a file and throw_error is set to true    
    """
    if can_be_directory:
        is_file = Path(filename).exists()
    else:
        is_file = Path(filename).is_file()
    if throw_error and not is_file:
        raise FileNotFoundError(f"{filename} is not a file.")
    return is_file


def get_file_type(filename: str, **kwargs) -> str:
    """
    Get the file type using the magic library.

    Args:
        filename (str): The filename to check
        kwargs (dict): Additional keyed parameters to the `from_file` method

    Returns:
        str: The file type. See libmagic documentation for supported types.
    """
    try:
        import magic
    except ImportError as e:
        LOGGER.fatal(e)
        exit()
    return magic.from_file(filename, **kwargs)


def get_file_extension(filename: str) -> str:
    """
    Get the extension for a file

    Args:
        filename (str): The file to get the extension for

    Returns:
        str: The file extension
    """
    return Path(filename).suffix

def get_resolved_path(path, av_relative: bool = False, return_as_str: bool = True) -> Union[str, Path]:
    """
    Get the fully resolved path to a specific file. If ``av_relative`` is set to true,
    the desired filename is relative to the ``av`` root subdirectory.

    Args:
        path (str): The path to get a fully resolved path from
        av_relative (bool): Whether the filepath is relative to the av subfolder. Defaults to False.
        return_as_str (bool): Returns the path as a string. Otherwise will return as a pathlib.Path object. Defaults to True

    Returns:
        Union[str, Path]: The fully resolved path as a string or Path object
    """
    path = Path(path)
    if av_relative:
        from av import __file__ as av_file
        path = Path(av_file).parent / path

    resolved_path = path.resolve()
    if return_as_str:
        return str(resolved_path)
    else:
        return resolved_path


def read_text(filename: str) -> str:
    """
    Read a file and return the text inside that file as a string

    Args:
        filename (str): The file to read

    Returns:
        str: The text inside filename
    """
    return Path(filename).read_text()

def search_upwards_for_file(filename: str) -> Path:
    """Search in the current directory and all directories above it 
    for a file of a particular name.

    Arg:
        filename (str): the filename to look for.

    Returns:
        Path: the location of the first file found or None, if none was found
    """
    d = Path.cwd()
    root = Path(d.root)

    while d != root:
        attempt = d / filename
        if attempt.exists():
            return attempt
        d = d.parent

    return None



"""
CLI command that handles interacting with the AV database.
"""

# rosbags imports
from rosbags.rosbag1 import Reader as ROS1Reader, Writer as ROS1Writer, ReaderError as ROS1ReaderError, WriterError as ROS1WriterError
from rosbags.rosbag2 import Reader as ROS2Reader, Writer as ROS2Writer, ReaderError as ROS2ReaderError, WriterError as ROS2WriterError
from rosbags.rosbag2.reader import decompress
from rosbags.rosbag2.connection import Connection as ROS2Connection
from rosbags.typesys import get_types_from_msg, register_types
from rosbags.serde import deserialize_cdr, ros1_to_cdr
from rosbags.serde.messages import get_msgdef
from rosbags.convert import convert, ConverterError

# Imports from avtoolbox
from avtoolbox.ros.messages import MessageType
from avtoolbox.utils.files import file_exists, get_file_type, get_file_extension, read_text, get_resolved_path, as_path, copy_file
from avtoolbox.utils.logger import LOGGER
from avtoolbox.utils.yaml_parser import YAMLParser

# General imports
from typing import NamedTuple, List, Tuple, Iterable, Any, Union
import sqlite3
import pickle
import pandas as pd
import shutil, os

# -------------
# Message Types
# -------------

def register_type(msg_file: Union['Path', str], name: str = None):
    """Registers a custom message type so that it can be read by ``rosbags``

    In order for custom messages to be read from ROS 2 bags, they must be registered with the ``rosbags``
    type system. This is do to an internal issue with ros2 bags that is going to fixed in an upcoming
    release (`see this issue <https://github.com/ros2/rosbag2/issues/782>`_).

    To register a message type, you must provide the file that the message is stored in (either with an 
    extension ``.msg`` or an extension ``.idl``), and the name of the the message (i.e. std_msgs/msg/Header).

    Args:
        msg_file (Union[Path, str]): The path the message definition is in. Must have an extension ``.msg`` or ``.idl``
        name (str, optional): The name of the custom message. Only optional if an ``idl`` file is passed.
    """
    ext = get_file_extension(msg_file)
    assert ext == '.msg' or ext == '.idl'

    text = as_path(msg_file).read_text()

    if ext == '.msg':
        assert name is not None
        register_types(get_types_from_msg(text, name))
    else:
        register_types(get_types_from_idl(text))

# --------
# Database
# --------

class AVDatabase:
    """
    This class allows you to programatically interact with the AV database.

    Typically, the AV CLI will be enough for pushing and pulling to and from the database.
    However, it may be desired to write scripts which pull data from the database. This class can be
    used in these instances.

    The AV database is contructed in a way that the `bag-database <https://swri-robotics.github.io/bag-database/>`_
    application can be used. Most of the restrictive descisions made for this package (i.e. using ROS 1 and ROS 2 bags)
    comes from requirements based on this package. The ``bag-database`` reads ros1 bag files from a regular directory
    and displays them, along with some additional features, in a web application. It is very helpful for development.

    The database, therefore, is really just a directory. The directory is filled with ros1 bag files and can either
    be managed locally or remotely with a url.

    .. warning::

        Currently only local databases are supported with the :meth:`~push` and :meth:`~pull` commands.
        
        .. raw:: html

           </div></div> 
    """
    def __init__(self, local_path: Union['Path', str]):
        self._local_path = as_path(local_path)

        if not self._local_path.exists():
            raise FileNotFoundError(f"Local path '{local_path}' doesn't exist.")

    def push(self, bag: Union['AVDataFile', str], keep: bool = True):
        """
        Push a database file to the AV database.

        The database is strictly made up of ros1 bags, so :meth:`~AVDataFile.to_ros1` will always
        be called. The file will then be copied (i.e. the original file still remains) to the database.

        Args:
            bag (Union[AVDataFile, str]): The bag file to push.
            keep (bool, optional): If False, will delete the file after pushing it.
        """
        if isinstance(bag, str):
            bag = AVDataFile(bag)

        copy_file(bag.path, self._local_path / as_path(bag.db_name))

        if not keep:
            shutil.rmtree(bag.path)

    def pull(self, name: str, dest: str = None, as_rosbag: bool = False) -> 'AVDataFile':
        """
        Pulls (downloads) a database file from the AV database

        Provided a name of the file to download, the file will be copied locally.
        
        Args:
            name (str): The name of the file to pull. Includes the extension.
            dest (str, optional): The destination of the file to be saved locally. If unset, will save with the same name is in the database.
            as_rosbag (bool, optional): If True, the pulled file will be saved as a ros1 bag.
        """
        name = as_path(name)
        if dest is None:
            dest = name.name
        copy_file(self._local_path / name, dest)

        data_file = AVDataFile(dest)
        if not as_rosbag:
            data_file.to_ros2()
            os.remove(dest)
        return data_file

    def ls(self, path: str = ""):
        """
        List the contents of the directory at the database root + ``path``.

        See `the man pages <https://man7.org/linux/man-pages/man1/ls.1.html>`_ for more information on ``ls``.

        Args:
            path(str, optional): Additional paths to navigate when running ``ls``.
        """
        import os
        return os.listdir(self._local_path / as_path(path))

# ---------
# Data File
# ---------

class AVDataFile:
    """
    This class represents a bag file that is stored locally.

    The file stored in the AV Database is a `rosbag <http://wiki.ros.org/rosbag>`_, or a 
    data storage method introduced in ROS 1. Although ROS 1 bags are stored in the database
    itself, either ROS 1 or ROS 2 bags can be used to push to the remote database. Conversions
    between the different data types when pushing to and pulling from the database.

    This class was created to abstract away the ROS 1/2 bag types from the database itself. In the future,
    when ROS 2 is more fully adopted, this class could simply use ROS 2 bags.
    It is desired to utilize ROS 2 bags now, but it is currently not possible to describe 
    custom message types in a ROS 2 bag, `see this issue <https://github.com/ros2/rosbag2/issues/782>`_.

    When in the AV database, a data file is idenfiable by it's name. The name has the following characteristics:
    - Begins with ``AV-``
    - Ends with the date the file was created in the format of ``mm-dd-YYYY-HH-MM-SS``

    For example, if a file was created on January 13th, 2022, at 9:00:00am, the database file will have the 
    following name: ``AV-01-13-2022-15-00-00``.

    .. note::

        The ``HH-MM-SS`` (i.e. hour-minute-second) is defined in terms of the Universal Coordinate Time (UTC).
        
        .. raw:: html

           </div></div> 

    ROS 1 bags are files with extension ``.bag``, i.e. ``AV-01-13-2022-15-00-00.bag``. ROS 2 bags are 
    directories, i.e. ``AV-01-13-2022-15-00-00/``.

    When pulling a file from the database, it *will* have a name. If you're working with a file locally and have
    not interacted with the database at all, it will generate a new name using the aforementioned structure.
    """

    def __init__(self, path: str):
        # Ensure the path actually exists
        self._path = path
        file_exists(self._path, throw_error=True, can_be_directory=True)

        # Generate the name for the file based on the start time of the bag (assumed to be time of creation)
        self._name = self._generate_name()

        self._type = _get_bag_version(path)

        self._reader = None

    def _generate_name(self):
        """Generates the name of the av data file

        *Should* be deterministic. Will grab the start time from the bag file (either ros1 or ros2)
        and use that as the creation data/time.

        As a reminder, the format of the name is ``AV-mm-dd-YYYY-HH-MM-SS``.
        """
        import datetime 

        with AVDataFileReader(self._path) as reader:
            # reader.start_time returns a time in UTC and includes nanoseconds
            # We don't need nanoseconds
            start_time = reader.start_time // 1e9

        # Generate the date/time portion of the name
        time_str = datetime.datetime.fromtimestamp(start_time).strftime("%m-%d-%Y-%H-%M-%S")

        # Prepend the time with AV-
        name = "AV-" + time_str
        LOGGER.info(f"Generated name is {name}")

        return name

    @property
    def path(self):
        """Path property.
        """
        return self._path

    @property
    def name(self):
        """Name property.

        As a reminder, the format of the name is ``AV-mm-dd-YYYY-HH-MM-SS``. It *should* be unique.
        """
        return self._name

    @property
    def db_name(self):
        """Database name property.

        The file name of the bag stored in the database. This differs from :attr:`name` in that it may include
        an extension (i.e. ``.bag``)
        """
        return self._name + '.bag' if self._type == 1 else ''

    @property
    def reader(self):
        """Get the reader object for this data file
        
        Returns:
           AVDataFileReader: The file reader

        Raises:
            RuntimeError: The reader hasn't been initialized yet. See :meth:`~__enter__`.
        """
        if self._reader is None:
            raise RuntimeError("The reader has not been opened yet.")

        return self._reader

    def to_ros1(self):
        """
        Converts this database file to a ros1 bag.

        If the bag that this object represents is already a ros1 bag, nothing is done (a warning will be made). 
        If it represents a ros2 bag, a conversion will be done to ros1.
        """
        if _get_bag_version(self._path) == 1:
            LOGGER.warn(f"'{self._path}' is already a ros1 bag. Doing nothing.")
        else:
            path = as_path(self._path)
            name = as_path(self._name + ".bag")
            if name.exists():
                LOGGER.error(f"'{name}' already exists. Doing nothing.")
                return
            convert(path, name)
            LOGGER.info(f"Successfully saved '{path}' as a ros1 bag to '{name}'")

    def to_ros2(self):
        """
        Converts this database file to a ros2 bag

        If the bag that this object represents is already a ros2 bag, nothing is done (a warning will be made). 
        If it represents a ros1 bag, a conversion will be done to ros2.
        """
        if _get_bag_version(self._path) == 2:
            LOGGER.warn(f"'{self._path}' is already a ros2 bag. Doing nothing.")
        else:
            path = as_path(self._path)
            name = as_path(self._name)
            if name.exists():
                LOGGER.error(f"'{name}' already exists. Doing nothing.")
                return
            convert(path, name)
            LOGGER.info(f"Successfully saved '{path}' as a ros2 bag to '{name}'")

    @staticmethod
    def combine(ros1_bag: str,
                ros2_bag: str,
                output_bag: str,
                ros1_topics: List[str] = [],
                ros2_topics: List[str] = [],
                ros1_types: List[MessageType] = [],
                ros2_types: List[MessageType] = []):
        """
        Combine command.

        Will combine a ros1 and ros2 bag. The output is a ros2 bag with both messages
        in the data with the respective timestamps. Additionally, to allow for easy parsing
        outside of ROS, a new table is created that describes the custom message types within the
        bag. When parsing, this information is used to register types for the rosbags python
        package.

        Args:
            ros1_bag        (str): The ros1 bag file to use in the combine step. ROS 1 uses files ending in .bag. Required.
            ros2_bag        (str): The ros2 bag file to use in the combine step. ROS 2 uses folders with a .db3 and metadata.yaml file. This should be a path to a folder. Required.
            output_bag      (str): The output ros2 bag file that contains the combined data. ROS 2 uses folders with a .db3 and metadata.yaml file. This should be a path to a folder. Required.
            ros1_topics     (List[str], optional): The topics to copy over from the ros1 bag. If empty, will copy all topics. Default is [] (empty).
            ros2_topics     (List[str], optional): The topics to copy over from the ros2 bag. If empty, will copy all topics. Default is [] (empty).
            ros1_types      (List[av.ros.messages.MessageType], optional): The custom message types to register when reading. Default is [] (will not register any custom message types).
            ros2_types      (List[av.ros.messages.MessageType], optional): The custom message types to register when reading. Default is [] (will not register any custom message types).

        Raises:
            KeyError: Raised if an unidentified message type is attempted to be parsed
        """
        LOGGER.info("Running combine command...")

        # Read through both bag files at the same time
        with AVDataFileReader(ros1_bag) as ros1_reader, AVDataFileReader(ros2_bag) as ros2_reader:
            # Contruct a connection list that is used to filter topics in the bag files
            ros1_conns = [x for x in ros1_reader.connections.values() if x.topic in ros1_topics]  # noqa
            ros2_conns = [x for x in ros2_reader.connections.values() if x.topic in ros2_topics]  # noqa

            # Create a generator for the ros1 and ros2 messages, respectively
            ros1_messages = ros1_reader.messages(connections=ros1_conns)
            ros2_messages = ros2_reader.messages(connections=ros2_conns)

            # Create a writer that will be used to write data to the rosbag file
            with AVDataFileWriter(output_bag) as writer:
                # If there any custom types, we will need to register them to read the bags
                # Also, to read them back easily later, we'll store the message types in the
                # bag itself
                types = ros1_types + ros2_types
                if types:
                    # Register the types
                    add_types = {}
                    for message in types:
                        msg_def = read_text(message.file)
                        add_types.update(get_types_from_msg(msg_def, message.name))
                    register_types(add_types)

                    # Create the table to store the message types for later
                    writer.add_message_types_table(add_types)

                # Remember the connections now
                # ros1 and ros2 connections are different and then the writer has it's own
                # Create a map from id/cid to the connection
                writer_conns = {}
                for conn in {**ros1_reader.connections, **ros2_reader.connections}.values():
                    writer_conn = writer.add_connection(conn.topic, conn.msgtype)
                    if isinstance(conn, ROS2Connection):
                        writer_conns[conn.id] = writer_conn
                    else:
                       writer_conns[conn.cid] = writer_conn

                # Now finally read the ros bags and store them in the new file
                # The insert order doesn't matter, the rosbags library will read the bag in ascending time order
                for i, (conn, timestamp, rawdata) in enumerate(ros1_messages):
                    try:
                        msgtype = writer_conns[conn.cid].msgtype
                        get_msgdef(msgtype)
                    except KeyError as e:
                        raise KeyError(f"{msgtype} isn't registered.")
                    writer.write(writer_conns[conn.cid], timestamp, ros1_to_cdr(rawdata, conn.msgtype))
                LOGGER.info(f"Inserted {i} messages from the ROS1 bag")
                for i, (conn, timestamp, rawdata) in enumerate(ros2_messages):
                    try:
                        msgtype = writer_conns[conn.id].msgtype
                        get_msgdef(msgtype)
                    except KeyError as e:
                        raise KeyError(f"{msgtype} isn't registered.")
                    writer.write(writer_conns[conn.id], timestamp, rawdata)
                LOGGER.info(f"Inserted {i} messages from the ROS2 bag")

    def open(self):
        """Open up the data file for reading.

        This is *not* a recommended method for reading the AV data file. Instead,
        you should use the contextmanager with :meth:`~__enter__`.

        It is not recommended because you must explicitly open *and* close the file. If unclosed,
        the file may become corrupt or other processes may be unable to open it.

        Usage:

        .. highlight:: python
        .. code-block:: python

            bagfile = "SOME_ROSBAG_FILE.bag"

            # Open the bag file
            file = AVDataFile(bagfile)
            file.open()

            for timestamp, connection, msg in file.reader:
                print(timestamp, msg)

            # Don't forget to close it!!
            file.close()

        Raises:
            AssertionError: If the reader is already open
        """
        assert self._reader is None

        self._reader = AVDataFileReader(self._path)
        self._reader.open()

    def close(self):
        """Close the opened data file.

        This is required to be called if :meth:`~open` is called explicitly. It is instead recommended to use
        the contextmanager with the :meth:`~__enter__` method.

        Raises:
            AssertionError: If the reader has not been opened yet
        """
        assert self._reader is not None

        self._reader.close()
        self._reader = None


    def __enter__(self):
        """Read the file when entering contextmanager.

        This is the preferred method of reading a AV database file.

        Examples:

        .. highlight:: python
        .. code-block:: python

            bagfile = "SOME_ROSBAG_FILE.bag"

            with AVDataFile(bagfile) as file:
                for timestamp, connection, msg in file.reader:
                    print(timestamp, msg)

        Alternatively, see :meth:`~open` and :meth:`~close`.
        """
        self.open()

        return self

    def __exit__(self, *args, **kwargs):
        """Close the file when entering contextmanager."""
        self.close()

        return False

# ----------------
# Data File Writer
# ----------------

class _AVROS1DataFileWriter(ROS1Writer):
    """
    Helper class to write to a ROS 1 bag.

    Is a simple wrapper around the :class:`rosbags.ros1.writer.Writer` class. Doesn't actually
    add any additional functionality.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class _AVROS2DataFileWriter(ROS2Writer):
    """
    Helper class to write to a ROS 2 bag.

    Is a simple wrapper around the :class:`rosbags.ros2.writer.Writer` class. Adds additional ability
    to add a custom table for message types.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add_message_types_table(self, types : 'rosbags.typesys.core.Typesdict'):
        """
        Creates a table to store metadata for the message types in the rosbag.

        One of the limitations for ros2 bags is that message types aren't defined within the file itself,
        so parsing it without knowledge of those types is not possible. We'll save the message types in the
        rosbag so we can use them when parsing it later. It will be stored simply as a pickle and will be
        unpickled when we read the bag file. 

        Args:
            types (rosbags.typesys.core.Typesdict):   the dict of types that rosbags.register_types expects
        """
        # Delete the messages table if it exists
        self.cursor.execute("DROP TABLE IF EXISTS message_types;")

        # Create a new table for the metadata
        sql = """
        CREATE TABLE IF NOT EXISTS message_types (
            types binary
        );
        """
        self.cursor.execute(sql)

        # Pickle the types dict and store it in the newly made table
        pdata = pickle.dumps(types, pickle.HIGHEST_PROTOCOL)
        sql = f"INSERT INTO message_types (types) VALUES(?)"
        self.cursor.execute(sql,(sqlite3.Binary(pdata),))

class AVDataFileWriter:
    """Simple wrapper class that allows the writing of ROS 1 *or* ROS 2 bags

    The writer classes do *not* append or overwrite existing bags. They will only write new bags.

    Args:
        path (Union[Path, str]): The path to write a file to.
        type (int): The ROS version type of the bag you want to write. Can be 1 or 2. If 0, the type will attempted to be inferred.
    """

    def __init__(self, path: 'Union[Path, str]', type: int = 0):
        self._writer = None
        self._path = path

        self._type = _get_bag_version(path)

    def open(self) -> 'Union[_AVROS1DataFileWriter, _AVROS2DataFileWriter]':
        """
        Method which will return either a :class:`~_AVROS1DataFileWriter` or a 
        :class:`~_AVROS2DataFileWriter`.

        Returns:
            Union[_AVROS1DataFileWriter, _AVROS2DataFileWriter]: The correct writer for the bag type.

        Raises:
            RuntimeError: If the bag already exists. The writer will only write new bags.
        """
        if self._type == 1:
            try:
                writer = _AVROS1DataFileWriter(self._path)
                writer.open()
                LOGGER.info(f"Opened '{self._path}' as a ROS 1 bag.")
                return writer
            except ROS1WriterError as e:
                LOGGER.error(e)
        elif self._type == 2:
            try:
                writer = _AVROS2DataFileWriter(self._path)
                writer.open()
                LOGGER.info(f"Opened '{self._path}' as a ROS 2 bag.")
                return writer
            except ROS2WriterError as e:
                LOGGER.error(e)

        raise RuntimeError(f"'{self._path}' already exists. AVDataFileWriter can only write new ros bags.")

    def close(self):
        """
        Method to close the writer.

        Raise:
            RuntimeError: If the writer was never opened.
        """
        if self._writer is None:
            raise RuntimeError(f"Writer was never opened.")

        self._writer.close()
        self._writer = None

    def __enter__(self):
        """Open ROS bag when entering contextmanager."""
        self._writer = self.open()
        return self._writer

    def __exit__(self, *args, **kwargs):
        """Close ROS bag when entering contextmanager."""
        self.close()
        return False 

# ----------------
# Data File Reader
# ----------------

class AVDataFileReader:
    """Simple wrapper class that allows the reading of ROS 1 *or* ROS 2 bags

    Args:
        path (Union[Path, str]): The path to write a file to.
    """

    def __init__(self, path: 'Union[Path, str]'):
        self._writer = None
        self._path = path

        self._type = _get_bag_version(path)

        self._reader = None

        # rosbags requires different ways of deserializing the bag files depending on the version
        # If it's a ros2 bag, just deserialize it
        # If it's a ros1 bag, we'll need to first serialize it to ros2 format (cdr), then deserialize it
        self._deserialize_func = deserialize_cdr if self._type == 2 else lambda raw, type: deserialize_cdr(ros1_to_cdr(raw, type), type)

    def convert_to_pandas_df(self) -> pd.DataFrame:
        """
        Convert the sqlite database to a pandas dataframe.

        Orders the database when reading by ascending order in terms of time.

        Example usage:

        .. highlight:: python
        .. code-block:: python

            from avtoolbox.db import AVDataFileReader

            bagfile = 'bag' # ros2 bag folder

            # Or you can use pandas
            with AVDataFileReader(bagfile) as reader:
                df = reader.convert_to_pandas_df()
                print(df)

        Returns:
            pandas.DataFrame: The ordered dataframe taken from the database.
        """
        # Read the messages from the generator into a pandas dataframe
        df = pd.DataFrame(self._reader.messages(), columns=['connections', 'timestamps', 'messages'])
        df = df[['timestamps', 'connections', 'messages']] # reorder the columns 
        
        # Deserialize the messages
        # TODO: Would ideally do this as we go. Possible?
        df['messages'] = df.apply(lambda r: self._deserialize_func(r['messages'], r['connections'].msgtype), axis=1)

        return df

    def open(self) -> 'Union[_AVROS1DataFileReader, _AVROS2DataFileReader]':
        """
        Method which will return either a :class:`~_AVROS1DataFileReader` or a 
        :class:`~_AVROS2DataFileReader`.

        Returns:
            Union[_AVROS1DataFileReader, _AVROS2DataFileReader]: The correct reader for the bag type.

        Raises:
            RuntimeError: If the bag already exists. The writer will only write new bags.
        """
        if self._type == 1:
            try:
                reader = _AVROS1DataFileReader(self._path)
                reader.open()
                LOGGER.info(f"Opened '{self._path}' as a ROS 1 bag.")
                self._reader = reader
            except ROS1ReaderError as e:
                LOGGER.error(e)
        elif self._type == 2:
            try:
                reader = _AVROS2DataFileReader(self._path)
                reader.open()
                LOGGER.info(f"Opened '{self._path}' as a ROS 2 bag.")
                self._reader = reader
            except ROS2ReaderError as e:
                LOGGER.error(e)

        if self._reader is None:
            raise RuntimeError(f"'{self._path}' doesn't exist or some other issue occurred.")

    def close(self):
        """
        Method to close the reader.

        Raise:
            RuntimeError: If the reader was never opened.
        """
        if self._reader is None:
            raise RuntimeError(f"Reader was never opened.")

        self._reader.close()
        self._reader = None

    def __enter__(self):
        """Open ROS bag when entering contextmanager."""
        self.open()
        return self

    def __exit__(self, *args, **kwargs):
        """Close ROS bag when entering contextmanager."""
        self.close()

    def __iter__(self) -> Iterable[Tuple[int, Union['ROS1Connection', ROS2Connection], Any]]:
        """
        Iterate method that returns a generator tuple.

        Example usage:

        .. highlight:: python
        .. code-block:: python

            from avtoolbox.db import AVDataFileReader

            bagfile = 'bag' # ros2 bag folder

            # You can use a generator
            with AVDataFileReader(bagfile) as reader:
                for timestamp, connection, msg in reader:
                    print(timestamp, msg)

        Yields:
            Tuple[int, Union[ROS1Connection, ROS2Connection], Any]: The timestamp for the message, the connection object for this message, and the deserialized message
        """
        LOGGER.debug(f"Reading from {str(self._path)} in ascending time...")

        for i, (connection, timestamp, rawdata) in enumerate(self._reader.messages()):
            yield timestamp, connection, self._deserialize_func(rawdata, connection.msgtype)
        return False 

    def __getattr__(self, attr):
        return getattr(self._reader, attr)
        

class _AVROS1DataFileReader(ROS1Reader, AVDataFileReader):
    """
    Helper class to read from a ROS 1 bag.

    Is a simple wrapper around the :class:`rosbags.ros1.reader.Reader` class. Doesn't actually
    add any additional functionality.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class _AVROS2DataFileReader(ROS2Reader, AVDataFileReader):
    """
    Helper class to read from a ROS 2 bag that was written by the av package

    It simply wraps the :class:`rosbags.ros2.reader.Reader` class. Adds additional functionality
    to read from a custom meta data table to register message types without needing to know
    the path to where they're stored.

    An example use case can be seen below:

    .. highlight:: python
    .. code-block:: python

        from avtoolbox.db import AVDataFileReader

        bagfile = 'bag' # ros2 bag folder

        # You can use a generator
        with AVDataFileReader(bagfile) as reader:
            for timestamp, connection, msg in reader:
                print(timestamp, msg)

        # Or you can use pandas
        with AVDataFileReader(bagfile) as reader:
            df = reader.convert_to_pandas_df()
            print(df)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._register_message_types()

    def _register_message_types(self):
        """
        Private method called by the constructor that registers messages.

        Using the :class:`~AVDataFileWriter` class, custom messages may be needed to parse the data types from
        either ROS1 or ROS2. As a result, using that class, we've stored some information regarding those
        message types in a separate sqlite table from the one holding all the ROS info. This method will
        parse that metadata table and register the additional types it finds.

        This method will work for rosbags not written by the :class:`~AVDataFileWriter` class, it
        just won't do anything if it can't find the message table. This is the same functionality if the
        writer didn't actually need to register any topics when it was writing.
        """

        for filepath in self.paths:
            with decompress(filepath, self.compression_mode == 'file') as path:
                conn = sqlite3.connect(f"file:{path}#immutable=1", uri=True)
                cursor = conn.cursor()

                # First, check if the message_types table exists
                sql = """
                SELECT count(name) FROM sqlite_master WHERE type='table' AND name='message_types';
                """
                c = cursor.execute(sql)

			    # Table will exist if the number of tables titled 'message_types' is 1	
                if c.fetchone()[0] == 1:
                    # Next, query for the actual types and then register their types
                    sql = """
                    SELECT * FROM message_types;
                    """
                    c = cursor.execute(sql)

                    # Register the types
                    for data in c:
                        register_types(pickle.loads(data[0]))


# ---------------------------
# Database Specific Utilities
# ---------------------------

def _get_bag_version(path: str) -> int:
    """Privattet helper method that guesses the bag version (1 or 2) given the path.

    ROS 1 bags are single files with extensions ``.bag``. ROS 2 bags are folders. We will simply check
    if the path is a file and assume that it is a ROS 1 bag, if so.

    Returns:
        int: 1 if the path is a rosbag (ROS 1), 2 if it is a ROS 2 bag
    """
    path = as_path(path)

    if path.is_file():
        type = 1
    elif path.exists():
        type = 2
    else:
        type = 1 if '.bag' in str(path) else 2
        
    return type

# -----------
# CLI Methods
# -----------

def _run_push(args):
    """
    Entrypoint for the `db push` command.

    The push command will copy local data files to the AV database. The AV database
    holds ros1 bags and is maintained as to allow [bag-database](https://swri-robotics.github.io/bag-database/)
    to read it.

    The `db push` command requires the local directory (see warning below) that the file will be pushed to.
    This is always the last positional argument. By default, it will push all ros1 or ros2 bags prefixed with 
    `AV-`. Otherwise, you may override this functionality by providing data files prior to the local directory
    path. See below for examples.

    ```bash
    # Push all data files in the current directory to /var/ros/data/
    av db push /var/ros/data/

    # Push just ros1.bag to the database
    av db push ros1.bag /var/ros/data/

    # Push just ros1.bag to the database
    av db push ros1.bag ros2bag/ /var/ros/data/

    # Push all bags in bags/ to the database
    av db push bags/* /var/ros/data/
    ```

    ```{warning}
    Currently only local databases are supported with :meth:`~push`.
    ```
    """
    LOGGER.debug("Running 'db push' entrypoint...")

    # Grab all the files
    files = []
    if len(args.files):
        files.extend(args.files)
    else:
        # If no files are passed, grab all the files in the current directory that start with "AV-"
        import glob

        av_files = glob.glob("AV-*")
        if len(av_files) == 0:
            LOGGER.fatal("No data files were provided and no bag files prefixed with 'AV-' where found in the current directory.")
            return
        files.extend(av_files)

    # Push each file
    if not args.dry_run:
        LOGGER.info(f"Connecting with database at '{args.local_path}'")
        db = AVDatabase(args.local_path)

        for file in files:
            if not file_exists(file):
                LOGGER.fatal(f"'{file}' is not a bag that can be pushed.")
                return

            LOGGER.info(f"Pushing '{file}'")
            db.push(file)


def _run_combine(args):
    """Command to combine a ROS 1 bag file and a ROS 2 bag into a single ROS 2 bag.

    The original motivation for this command is that the [mocap optitrack](http://wiki.ros.org/mocap_optitrack)
    package was only supported in ROS 1, but the control stack we were using was in ROS 2. It is desired that
    the bagfiles are combined, timestamp corrected, so that replaying the data includes the ground truth of the
    robot.

    Example usage:

    ```bash
    av -vv db combine -c configuration.yml
    ```

    `configuration.yml`:

    ```yaml
    output:
      file: output

    rosbag1:
      file: ros1.bag
      topics:
        - /mocap_node/av/pose

    rosbag2:
      file: ros2bag/
      messages:
        vehicle_input:
          file: data/VehicleInput.msg
          name: custom_msgs/msg/VehicleInput
    ```
    """
    LOGGER.debug("Running 'db combine' entrypoint...")

    # Parse the YAML config file first
    if not file_exists(args.config, throw_error=False) and (not args.ros1bag and not args.ros2bag and not args.output):
        raise ValueError(f"combine command requires either a config file or all three of the following: a ros1bag, ros2bag and output path name")
    yaml_parser = YAMLParser(args.config)

    # Get the ros1 and ros2 bag files
    ros1bag = yaml_parser.get('rosbag1', 'file', default=args.ros1bag)
    ros2bag = yaml_parser.get('rosbag2', 'file', default=args.ros2bag)

    # Get the available topics
    ros1_topics = yaml_parser.get('rosbag1', 'topics', default=args.ros1_topics)
    ros2_topics = yaml_parser.get('rosbag2', 'topics', default=args.ros2_topics)

    # Get the available messages
    ros1_messages = [MessageType(file=d["file"], name=d["name"]) for d in yaml_parser.get('rosbag1', 'messages', default=args.ros1_messages).values()]
    ros2_messages = [MessageType(file=d["file"], name=d["name"]) for d in yaml_parser.get('rosbag2', 'messages', default=args.ros2_messages).values()]

    # Get the output filename
    output = yaml_parser.get('output', 'file', default=args.output)

    # Do some checks to make sure the files exists/filetypes are correct
    assert file_exists(ros1bag, throw_error=True)

    # Run the combine command
    AVDataFile.combine(ros1bag, ros2bag, output, ros1_topics=ros1_topics, ros2_topics=ros2_topics, ros1_types=ros1_messages, ros2_types=ros2_messages)


def _run_read(args):
    """Command to read a AV database file.

    The AV database files are unique in that they allow you to read a ROS 2 bag with custom message types.
    With ROS 2 Galatic, [this is not possible](https://github.com/ros2/rosbag2/issues/782), hence the need to
    do this ourselves. To implement this, the `rosbags` package ([documentation](https://ternaris.gitlab.io/rosbags/))
    is utilize to convert between ROS 1 and ROS 2 bags without using either as an actual dependency. By default,
    ROS 2 bags are stored in an sqlite database and to inform the reader of the bag of the custom message types,
    a new sqlite table is added with the custom definition (stored as a pickle that `rosbags` can use to register
    the message types).

    The advantage of adding our own table and keeping to the sqlite structure is that you can still replay the ros2
    bag as if it was recorded normally and unaffected (assuming you have sourced a workspace with the custom msg
    types).

    This command is simply a debug tool for reading a AV database file.

    Example usage:

    ```bash
    av -vv db read input

    # OR

    av -vv db read configuration.yaml
    ```

    `configuration.yml`:

    ```yaml
    input:
      file: input
      messages:
        vehicle_input:
          file: data/VehicleInput.msg
          name: custom_msgs/msg/VehicleInput
    ```

    """
    LOGGER.debug("Running 'db read' entrypoint...")

    # Parse the YAML config file first
    assert file_exists(args.config, throw_error=True)
    yaml_parser = YAMLParser(args.config)

    # Get the input filename
    input = yaml_parser.get('input', 'file', default=args.input)

    # Register the types that we need
    ros2_messages = [MessageType(file=d["file"], name=d["name"]) for d in yaml_parser.get('input', 'messages', default=[]).values()]
    register_type("data/VehicleInput.msg", "custom_msgs/msg/VehicleInput")

    # Read and print out the data
    # TODO: Make this more useful. Metadata (num messages, time, etc.)?
    with AVDataFileReader(input) as reader:
        for i, (timestamp, topic, msg) in enumerate(reader):
            print(timestamp, topic)

def _init(subparser):
    """Initializer method for the `db` entrypoint

    This entrypoint provides easy manipulation of the AV database. The database is simply organized
    in a directory located either on a remote system or locally. The directory holds ROS 1 bags. At the time of
    creation, it was desired to have a way to parse the database in a rich GUI environment and be able to visually
    inspect topics, bags, and other relevant data. To easily do this, the [bag-database](https://github.com/swri-robotics/bag-database)
    application is used. When written, [ROS 2 bags](https://github.com/ros2/rosbag2/issues/782) lack the ability
    to contain information regarding custom message types (the next release of ROS in May, 2022, should implement this feature).

    This cli tool will therefore then provide a way to easily convert ROS 2 bags to ROS 1 to store in the AV
    database. Commands will simplify the "pushing" and "pulling" to and from the database. Furthermore, additional
    tools have been written such as combining ROS 1 and ROS 2 bags into a single ROS 2 bag.

    In the future, the workflow can be adjusted to use ROS 2 bags once `bag-database` implements it, assuming the
    next release of ROS 2 and ros2 bags supports it.
    """
    # Create some entrypoints for additional commands
    subparsers = subparser.add_subparsers(required=False)

    # Push subcommand
    # Convert local ros2 bag to a rosbag (ROS 1) and then push it to the AV Database
    push = subparsers.add_parser("push", description="Push a bag file to the AVDatabase.")
    push.add_argument("files", nargs="*", help="The files to push. If none are provided, all AV- prefixed files are selected.")
    push.add_argument("local_path", help="The local path to the directory of the database.")
    push.set_defaults(cmd=_run_push)

    # Combine subcommand
    # Used to combine ros bag files
    combine = subparsers.add_parser("combine", description="Combine a ROS1 and ROS2 bag by matching timestamps.")
    combine.add_argument("-r1", "--ros1bag", help="ROS1 bag file. Will be overridden if present in the yaml.", default="")
    combine.add_argument("-r2", "--ros2bag", help="ROS2 bag file. Will be overridden if present in the yaml.", default="")
    combine.add_argument("-o", "--output", help="The output filename for the sqlite3 database", default="")
    combine.add_argument("-c", "--config", help="YAML file that defines the conversion process", default="")
    combine.add_argument("--ros1_topics", help="The ros1 topics", default=[])
    combine.add_argument("--ros2_topics", help="The ros2 topics", default=[])
    combine.add_argument("--ros1_messages", help="The ros1 messages", default={})
    combine.add_argument("--ros2_messages", help="The ros2 messages", default={})
    combine.set_defaults(cmd=_run_combine)

    # Read subcommand
    # Used to read ros bag files
    read = subparsers.add_parser("read", description="Read the custom av sqlite database files.")
    read.add_argument("config", help="YAML file that defines the read process")
    read.add_argument("-i", "--input", help="The database file to read")
    read.set_defaults(cmd=_run_read)



"""A setuptools based setup module."""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')


def parse_requirements():
    with open('requirements.txt') as f:
        required = f.read().splitlines()

    return required

def create_version():
    from setuptools_scm.version import get_local_dirty_tag

    def clean_scheme(version):
        return get_local_dirty_tag(version) if version.dirty else '+clean'

    return {'local_scheme': clean_scheme}

setup(
    name='avtoolbox',  # Required
    use_scm_version=create_version,
    description='Autonomous Vehicle Toolbox',  # Optional
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    author='UW Simulation Based Engineering Lab',  # Optional
    license="BSD3",
    packages=find_packages(),  # Required
    include_package_data=True,
    python_requires='>=3.8.2, <4',
    install_requires=parse_requirements(),  # Optional
    url="https://projects.sbel.org/avtoolbox",
    entry_points={  # Optional
        'console_scripts': [
            'av=avtoolbox._av_base:_main',
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    project_urls={  # Optional
        "Homepage": "https://projects.sbel.org/avtoolbox",
        "Bug Reports": "https://github.com/uwsbel/avtoolbox/issues",
        "Source Code": "https://github.com/uwsbel/avtoolbox/",
        "Our Lab!": "https://sbel.wisc.edu",
    },
)



"""
CLI command that handles working with the AV development environment
"""

# Imports from av
from avtoolbox.utils.logger import LOGGER
from avtoolbox.utils.yaml_parser import YAMLParser
from avtoolbox.utils.files import search_upwards_for_file, read_text

# Other imports
from python_on_whales import docker, DockerClient, exceptions as docker_exceptions
import yaml
import pkgutil
import os
import tempfile
import socket
import pathlib

def _run_env(args):
    """
    Entrypoint for the `dev env` command.

    The `env` command essentially wraps `docker-compose` to automatically build, spin-up, attach, and
    tear down the AV development environment.

    The command is completely redundant; it simply combines all the building, starting, attaching, and destroying
    into a single command. It allows users to quickly start and attach to the AV development environment based
    on the docker-compose file.

    There are four possible options that can be used using the `env` subcommand:
    `build`, `up`, `down`, and `attach`. For example, if you'd like to build the container, you'd run 
    the following command:

    ```bash
    av dev env --build
    ```

    If you'd like to build, start the container, then attach to it, run the following command:

    ```bash
    av dev env --build --up --attach
    # OR
    av dev env -b -u -a
    # OR
    av dev env -bua
    ```

    If no arguments are passed, this is equivalent to the following command:

    ```bash
    av dev env --up --attach
    ```

    If desired, pass `--down` to stop the container. Further, if the container exists and changes are
    made to the repository, the container will _not_ be built automatically. To do that, add the 
    `--build` argument.

    ```{note}
    `av dev` is an alias for `av dev env`. `env` can therefore be omitted for brevity.
    </div></div>
    ```
    """
    LOGGER.info("Running 'dev env' entrypoint...")

    # Check docker-compose is installed
    LOGGER.debug("Checking if 'docker compose' (V2) is installed...")
    if not docker.compose.is_installed():
        LOGGER.fatal("The command 'docker compose' is not installed. See http://projects.sbel.org/avtoolbox/tutorials/using_the_development_environment.html for more information.")
        return
    LOGGER.debug("'docker compose' (V2) is installed.")

    # Check that the .avtoolbox.conf file is located in this directory or any parent directories
    # This file should be at the root of any avtoolbox compatible stack
    LOGGER.info("Searching for '.avtoolbox.yml'...")
    conf = search_upwards_for_file('.avtoolbox.yml')
    if conf is None:
        LOGGER.fatal("No .avtoolbox.yml file was found in this directory or any parent directories. Make sure you are running this command in an avtoolbox compatible repository.")
        return
    root = conf.parent
    LOGGER.info(f"Found '.avtoolbox.yml' at {conf}.")

    # Read the conf file 
    # The conf file _can_ contain the following information (none required):
    # 
    # project: <str> (required)
    #   This is the name of the project that the docker containers will be named after
    #   The main development environment will be called {project}-dev and so on.
	#
    # user: <dict> (optional)
    #   name: <str> (optional)
    #       The user to create in the container. If not set, will grab the current user.
    #
    #   uid: <int> (optional)
    #       The user id to use. If not set, will grab it from the current user. (NOT USED)
    #
    #   gid: <int> (optional)
    #       The group id to use. If not set, will grab it from the current user. (NOT USED)
    #
    # dev: <dict> (optional)
    #   scripts: <path> (optional)
    #     This is the directory that holds any docker specific files (like custom dockerfiles or
    #     packages/scripts/dependencies that needed to be used when building the docker image).
    #     The path is relative to the .avtoolbox.yml file.
    #
    #   apt-dependences: <list> (optional)
    #       A list of requirements that should be installed in the image through apt.
    #
    #   pip-requirements: <list> (optional)
    #       A list of requirements that should be installed in the image through pip.
    #
    #   ports: <list> (optional)
    #       A list of ports to be exposed.
    #
    # vnc: <dict> (optional)
    #   novnc_port: <int> (optional)
    #       The port that novnc is displayed on. Defaults to 8080.
    #
    #   vnc_port: <int> (optional)
    #       The port that vnc is displayed on. Defaults to 5900.
    #
    #   password: <str> (optional)
    #       The password to use for vnc. Defaults to the project name.
    #
    # ros: <dict> (optional)
    #   workspace: <path> (optional)
    #     This is the directory which holds the ROS 2 workspace code
    #     The path is relative to the .avtoolbox.yml file.
    #
    #   distro: <str> (optional)
    #       The ros 2 distro to use. Must be compatible with osrf/ros distrobutions
    #       on DockerHub. Defaults to galactic.

    LOGGER.debug("Parsering '.avtoolbox.yml' file.")
    conf_parser = YAMLParser(conf)

    def get_attr(*args, default=None):
        if not conf_parser.contains(*args):
            LOGGER.debug(f"'{'.'.join(args)}' wasn't found in the .avtoolbox.yml file. Assuming '{'.'.join(args)}' to be '{default}'")
            ret = default
        else:
            ret = conf_parser.get(*args)
        return ret

    def read_data(*args):
        path = os.path.realpath(os.path.join(__file__, "..", *args))
        return read_text(path), path

    project = get_attr("project")
    if project is None:
        LOGGER.fatal(f"'project' must be provided in '.avtoolbox.yml'. Wasn't found, cannot continue.")
        return
    elif any(c.isupper() for c in project):
        LOGGER.fatal(f"'project' is set to '{project}' which is not allowed since it has capital letters. Please choose a name with only lowercase.")
        return

    username = get_attr("user", "name", default=os.getlogin())
    uid = get_attr("user", "uid", default=os.getuid())
    gid = get_attr("user", "gid", default=os.getgid())

    # Docker
    dev_dockerfile, dev_dockerfile_path = read_data("docker", "dev", "dev.dockerfile")
    apt_dependencies = ' '.join(get_attr("dev", "apt-dependencies", default=""))
    pip_requirements = ' '.join(get_attr("dev", "pip-requirements", default=""))
    scripts = get_attr("dev", "scripts", default=os.path.join("docker", "dev", "scripts"))
    ports = get_attr("dev", "ports", default=[])

    # ROS
    workspace = get_attr("ros", "workspace", default="workspace")
    distro = get_attr("ros", "distro", default="galactic")
    
    # VNC
    novnc_port = get_attr("vnc", "novnc_port", default="8080")
    vnc_port = get_attr("vnc", "vnc_port", default="5900")
    vnc_password = get_attr("vnc", "password", default=project)
    vnc_dockerfile, vnc_dockerfile_path = read_data("docker", "vnc", "vnc.dockerfile")

    # The docker containers are generated from a docker-compose.yml file
    # Grab the shipped configuration files that come with the toolbox package
    docker_compose, docker_compose_path = read_data("docker", "docker-compose.yml")
    docker_compose_fmt = {
        "PROJECT": project,
        "ROOT": root,
        "DEV.DOCKERFILE": dev_dockerfile_path,
        "VNC.DOCKERFILE": vnc_dockerfile_path,
        "USERNAME": username,
        "USER": f"{uid}:{gid}",
        "SCRIPTS": scripts,
        "APT-DEPENDENCIES": apt_dependencies,
        "PIP-REQUIREMENTS": pip_requirements,
        "NOVNC_PORT": novnc_port,
        "VNC_PORT": vnc_port,
        "VNC_PASSWORD": vnc_password,
        "WORKSPACE": workspace,
        "ROSDISTRO": distro,
    }

    for k,v in docker_compose_fmt.items():
        k = f"${{{k}}}"
        v = str(v)
        docker_compose = docker_compose.replace(k, v)
    docker_compose_yml = yaml.load(docker_compose, Loader=yaml.SafeLoader)

    # If no command is passed, start up the container and attach to it
    cmds = [args.build, args.up, args.down, args.attach] 
    if all(not c for c in cmds):
        args.up = True
        args.attach = True

    # Complete the arguments
    if not args.dry_run:
        def _is_port_in_use(port: int) -> bool:
            """Helper function to check if a port is currently in use."""
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                return s.connect_ex(('localhost', port)) == 0

        
        # We need to provide the ability to dynamically change the docker-compose file without it being in
        # version control. Meaning, for instance, if someone has an existing program using port 5900, we can
        # overwrite this value and give a new one. To accomplish this, we'll create a temporary file,
        # make any necessary edits to the yaml which is now in the temp file, run the commands we need,
        # then delete the temp file

        # Do this in a try/except so that the temp file will always be deleted
        try:

            # Create the temporary file and don't let it delete automatically yet
            tmp = tempfile.NamedTemporaryFile(delete=False)

            # Write the config to the temporary yaml file (first time)
            with open(tmp.name, 'w') as yaml_file:
                yaml.dump(docker_compose_yml, yaml_file)

            # Set the compose file to the temporary file
            client = DockerClient(compose_files=[tmp.name])
            config = client.compose.config(return_json=True)

            # Get main dev name
            if not any('dev' in service for service in config['services']):
                LOGGER.fatal(f"The docker-compose.yml configuration file doesn't contain a service that has 'dev' in it. Make sure you're running in this the correct place.")
                return

            for service_name, service in config['services'].items():
                if 'dev' in service_name:
                    dev_name = service['container_name']
                    break

            if args.down:
                LOGGER.info(f"Tearing down...")
                client.compose.down()

            if args.build:
                LOGGER.info(f"Building...")
                client.compose.build()

            if args.up:
                LOGGER.info(f"Spinning up...")

                # Ignore all running services
                config['services'] = {name : service for name, service in config['services'].items() if len(docker.container.list(filters={"name": service["container_name"]})) == 0 }

                # If all the services are already running, don't try to spin up
                if len(config['services']) == 0:
                    LOGGER.warn(f"No services need to be initialized. Turning off 'up'. If you didn't explicitly call 'up', this may be safely ignored.")
                    args.up = False

                for service_name, service in config['services'].items():

                    # For each port in each service, make sure they map to available ports
                    # If not, increment the published port by one. Only do this 5 times. If a port can't be found,
                    # stop trying.
                    LOGGER.debug("Checking if any host ports are already in use for service '{service_name}'.")

                    # Add any custom ports
                    if 'dev' in service_name:
                        if 'ports' in service:
                            service['ports'].extend(ports)
                        else:
                            service['ports'] = ports

                    if not 'ports' in service:
                        LOGGER.debug(f"'{service_name}' has no ports mapped. Continuing to next service...")
                        continue

                    for port in service['ports']:
                        for i in range(5):
                            published = port['published'] if isinstance(port, dict) else int(port.split(":")[0])
                            target = port['target'] if isinstance(port, dict) else port.split(":")[1]
                            if _is_port_in_use(published):
                                LOGGER.warn(f"Tried to map container port '{target}' to host port '{published}' for service '{service_name}', but it is in use. Trying again with '{published + 1}'.")
                                if isinstance(port, str):
                                    port = f"{published + 1}:{target}"
                                else:
                                    port['published'] += 1
                            else:
                                break

                # Write the config again to solidify any changes
                with open(tmp.name, 'w') as yaml_file:
                    yaml.dump(config, yaml_file)

                client.compose.up(detach=True)

            if args.attach:
                LOGGER.info(f"Attaching...")
                try:
                    usershell = [e for e in client.container.inspect(dev_name).config.env if "USERSHELL" in e][0]
                    shellcmd = usershell.split("=")[-1]
                    shellcmd = [shellcmd, "-c", f"{shellcmd}"]
                    print(client.execute(dev_name, shellcmd, interactive=True, tty=True))
                except docker_exceptions.NoSuchContainer as e:
                    LOGGER.fatal(f"The containers have not been started. Please run again with the 'up' command.")
        except docker_exceptions.DockerException as e:
            msg = str(e)
            if 'Error response from daemon:' in msg:
                msg = msg.split('Error response from daemon:')[1][:-3]
            LOGGER.error(f"Docker command raised exception: {msg}")
        finally:
            tmp.close()
            os.unlink(tmp.name)

def _init(subparser):
    """Initializer method for the `dev` entrypoint

    This entrypoint provides easy access to the AV development environment. The dev environment
    leverages [Docker](https://docker.com) to allow interoperability across operating systems. `docker-compose`
    is used to build, spin up, attach, and tear down the containers. The `dev` entrypoint will basically wrap
    the `docker-compose` commands to make it easier to customize the workflow to work best for AV.

    The primary container, titled `dev` in the `docker-compose.yml` file in the 
    [`av` github](https://github.com/uwsbel/avtoolbox), has [ROS 2](https://docs.ros.org/en/galactic/index.html)
    pre-installed. The software stack for the AV vehicle utilizes ROS 2 and will use
    the same container that is used for development. 

    Additional containers may be provided to allow for GUI windows or run simulations.
    """
    LOGGER.debug("Initializing 'dev' entrypoint...")

    def _add_dev_commands(parser):
        parser.add_argument("-b", "--build", action="store_true", help="Build the env.", default=False)
        parser.add_argument("-u", "--up", action="store_true", help="Spin up the env.", default=False)
        parser.add_argument("-d", "--down", action="store_true", help="Tear down the env.", default=False)
        parser.add_argument("-a", "--attach", action="store_true", help="Attach to the env.", default=False)
        parser.set_defaults(cmd=_run_env)

    # Add a base 
    _add_dev_commands(subparser)

    # Create some entrypoints for additinal commands
    subparsers = subparser.add_subparsers(required=False)

    # Subcommand that can build, spin up, attach and tear down the dev environment
    env = subparsers.add_parser("env", description="Command to simplify usage of the docker-based development workflow. Basically wraps docker-compose.")
    _add_dev_commands(env)

